# Copyright (c) 2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import unittest

from tellive.livemessage import LiveMessageToken


class Test(unittest.TestCase):
    def assert_invalid(self, token):
        self.assertIs(token.value, None)

    def assert_int(self, token, expected):
        self.assertIs(type(token.value), int)
        self.assertEqual(token.value, expected)

    def assert_string(self, token, expected):
        self.assertIs(type(token.value), str)
        self.assertEqual(token.value, expected)

    def assert_base64(self, token, expected):
        self.assertIs(type(token.value), bytes)
        self.assertEqual(token.value, expected)

    def assert_list(self, token, expected):
        self.assertIs(type(token.value), list)
        self.assertListEqual(token.value, expected)

    def assert_dict(self, token, expected):
        self.assertIs(type(token.value), dict)
        self.assertDictEqual(token.value, expected)

    def test_empty_token(self):
        self.assert_invalid(LiveMessageToken())

    def test_int(self):
        token = LiveMessageToken(10)
        self.assert_int(token, 10)
        self.assertEqual(token.serialize(), b'iAs')

    def test_string(self):
        a_ring = b'\xc3\xa5'.decode('utf-8')
        token = LiveMessageToken("foobar123456" + a_ring)
        self.assert_string(token, "foobar123456" + a_ring)
        self.assertEqual(token.serialize(), b'E:foobar123456\xc3\xa5')

    def test_base64(self):
        token = LiveMessageToken(b'foobar123456')
        self.assert_base64(token, b'foobar123456')
        self.assertEqual(token.serialize(), b'u10:Zm9vYmFyMTIzNDU2')

    def test_empty_list(self):
        token = LiveMessageToken([])
        self.assert_list(token, [])
        self.assertEqual(token.serialize(), b'ls')

    def test_list(self):
        token = LiveMessageToken([1, "foo"])
        self.assert_list(token, [1, "foo"])
        self.assertEqual(token.serialize(), b'li1s3:foos')

    def test_empty_dict(self):
        token = LiveMessageToken({})
        self.assert_dict(token, {})
        self.assertEqual(token.serialize(), b'hs')

    def test_dict(self):
        token = LiveMessageToken({'1': "foo", '2': 3})
        self.assert_dict(token, {'1': "foo", '2': 3})
        # The order of keys are undefined
        serialized = token.serialize()
        if serialized != b'h1:13:foo1:2i3ss':
            self.assertEqual(serialized, b'h1:2i3s1:13:foos')

    def test_dict_in_list_in_dict(self):
        token = LiveMessageToken({'a': [{'b': 1}, 2]})
        self.assertIs(type(token.value['a']), list)
        self.assertEqual(token.serialize(), b'h1:alh1:bi1ssi2sss')

    def test_deserialize_empty(self):
        self.assertIs(None, LiveMessageToken.deserialize(b'')[0])

    def test_deserialize_invalid(self):
        self.assertIs(None, LiveMessageToken.deserialize(b'i')[0])
        self.assertIs(None, LiveMessageToken.deserialize(b'l')[0])
        self.assertIs(None, LiveMessageToken.deserialize(b'h')[0])
        self.assertIs(None, LiveMessageToken.deserialize(b'hi1s')[0])
        self.assertIs(None, LiveMessageToken.deserialize(b'u')[0])
        self.assertIs(None, LiveMessageToken.deserialize(b'u7:YWxpdmU')[0])

    def test_deserialize_int(self):
        (token, rest) = LiveMessageToken.deserialize(b'i1A2s')
        self.assert_int(token, 0x1a2)
        self.assertFalse(rest)

    def test_deserialize_empty_list(self):
        (token, rest) = LiveMessageToken.deserialize(b'ls')
        self.assert_list(token, [])
        self.assertFalse(rest)

    def test_deserialize_list(self):
        (token, rest) = LiveMessageToken.deserialize(b'li1siAss')
        self.assert_list(token, [1, 10])
        self.assertFalse(rest)

    def test_deserialize_empty_dict(self):
        (token, rest) = LiveMessageToken.deserialize(b'hs')
        self.assert_dict(token, {})
        self.assertFalse(rest)

    def test_deserialize_dict(self):
        (token, rest) = LiveMessageToken.deserialize(b'h1:13:foo1:26:foobars')
        self.assert_dict(token, {'1': "foo", '2': "foobar"})
        self.assertFalse(rest)

    def test_deserialize_dict_with_list(self):
        (token, rest) = LiveMessageToken.deserialize(
            b'h1:1li1Es1:Alss1:2h1:3hsss')
        self.assert_dict(token, {'1': [0x1e, "A", []], '2': {'3': {}}})
        self.assertFalse(rest)

    def test_deserialize_empty_base64(self):
        (token, rest) = LiveMessageToken.deserialize(b'u0:')
        self.assert_base64(token, b'')
        self.assertFalse(rest)

    def test_deserialize_base64(self):
        (token, rest) = LiveMessageToken.deserialize(b'u8:YWxpdmU=')
        self.assert_base64(token, b'alive')
        self.assertFalse(rest)

    def test_deserialize_empty_string(self):
        (token, rest) = LiveMessageToken.deserialize(b'0:')
        self.assert_string(token, "")
        self.assertFalse(rest)

    def test_deserialize_string(self):
        (token, rest) = LiveMessageToken.deserialize(b'10:123456789abcdefP')
        self.assert_string(token, "123456789abcdefP")
        self.assertFalse(rest)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

from .livemessage import LiveMessage

import http.client as http
import socket
import ssl
import xml.parsers.expat as expat


class TellstickLiveClient(object):
    # dict(supportedMethods)
    SUBJECT_REGISTERD = "registered"
    # dict(uuid, url)
    SUBJECT_NOT_REGISTERED = "notregistered"
    # dict(id, action, [value, ACK])
    SUBJECT_COMMAND = "command"
    # <no parameters>
    SUBJECT_PONG = "pong"
    # <no parameters>
    SUBJECT_DISCONNECT = "disconnect"

    def __init__(self, public_key, private_key, debug=False):
        super(TellstickLiveClient, self).__init__()
        self.socket = None
        self.public_key = public_key
        self.private_key = private_key
        self.hash_method = "sha1"
        self.debug = debug

    def ssl_context(self):
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        context.verify_mode = ssl.CERT_REQUIRED
        context.set_default_verify_paths()
        return context

    def servers(self, server='api.telldus.com', port=http.HTTPS_PORT):
        """Fetch list of servers that can be connected to.

        :return: list of (address, port) tuples
        """
        conn = http.HTTPSConnection(server, port, context=self.ssl_context())
        conn.request('GET', "/server/assign?protocolVersion=2")

        response = conn.getresponse()
        if response.status != http.OK:
            raise ConnectionError("could not connect to {}:{}: {} {}".format(
                    server, port, response.status, response.reason))

        servers = []

        def extract_servers(name, attributes):
            if name == "server":
                servers.append((attributes['address'],
                                int(attributes['port'])))

        parser = expat.ParserCreate()
        parser.StartElementHandler = extract_servers
        parser.ParseFile(response)

        return servers

    def connect(self, address, timeout=5):
        sock = self.ssl_context().wrap_socket(
            socket.socket(socket.AF_INET, socket.SOCK_STREAM))
        sock.settimeout(timeout)
        sock.connect(address)
        ssl.match_hostname(sock.getpeercert(), address[0])
        self.socket = sock

    def connect_to_first_available_server(self, **kwargs):
        for server in self.servers():
            try:
                self.connect(server, **kwargs)
                return server
            except:
                pass
        raise ConnectionError("could not connect to any available server")

    def disconnect(self):
        if self.socket:
            self.socket.close()
            self.socket = None

    def send_message(self, message):
        envelope = message.create_signed_message(
            self.private_key, self.hash_method)

        if self.debug:
            print("-> ", envelope.serialize())
        self.socket.write(envelope.serialize())

    def receive_message(self):
        data = self.socket.read(1024)
        if self.debug:
            print("<- ", data)

        envelope = LiveMessage.deserialize(data)
        if not envelope.verify_signature(self.private_key, self.hash_method):
            raise ValueError("Signature verification failed")

        return LiveMessage.deserialize(envelope.parameter(0).encode('utf-8'))

    def register(self, version, uuid=""):
        message = LiveMessage("Register")
        message.append({'key': self.public_key, 'uuid': uuid,
                        'hash': self.hash_method})
        message.append({'protocol': 2, 'version': version,
                        'os': "linux", 'os-version': "unknown"})
        self.send_message(message)

    def ping(self):
        message = LiveMessage("Ping")
        self.send_message(message)

    def acknowledge(self, cookie):
        message = LiveMessage("ACK")
        message.append(cookie)
        self.send_message(message)

    def report_devices(self, devices, supported_methods):
        dev_list = []
        for device in devices:
            dev = {'id': device.id, 'name': device.name}
            dev['methods'] = device.methods(supported_methods)
            dev['state'] = device.last_sent_command(supported_methods)
            dev['stateValue'] = device.last_sent_value()
            dev_list.append(dev)
        message = LiveMessage("DevicesReport")
        message.append(dev_list)
        self.send_message(message)

    def report_device_event(self, device_id, method, data):
        message = LiveMessage("DeviceEvent")
        message.append(device_id)
        message.append(method)
        message.append(data)
        self.send_message(message)

    def _sensor(self, sensor):
        value_list = []
        for datatype in sensor.DATATYPES.values():
            if not sensor.has_value(datatype):
                continue
            value = sensor.value(datatype)
            value_list.append({'type': datatype, 'value': value.value,
                               'lastUp': value.timestamp})
            s = {'protocol': sensor.protocol, 'model': sensor.model,
                 'sensor_id': sensor.id}
        return s, value_list

    def report_sensors(self, sensors):
        sensor_list = []
        for sensor in sensors:
            s, value_list = self._sensor(sensor)
            s['name'] = ''
            sensor_list.append([s, value_list])

        message = LiveMessage("SensorsReport")
        message.append(sensor_list)
        self.send_message(message)

    def report_sensor_values(self, sensor):
        s, value_list = self._sensor(sensor)
        message = LiveMessage("SensorEvent")
        message.append(s)
        message.append(value_list)
        self.send_message(message)



# Copyright (c) 2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import base64
import hashlib


class LiveMessageToken(object):
    def __init__(self, value=None):
        super(LiveMessageToken, self).__init__()

        if type(value) in (int, str, bytes, list, dict):
            self.value = value
        else:
            self.value = None

    def serialize(self):
        """Serialize the token and return it as bytes."""
        if type(self.value) == int:
            return "i{:X}s".format(self.value).encode('ascii')

        if type(self.value) == str:
            value = self.value.encode('utf-8')
            return "{:X}:".format(len(value)).encode('ascii') + value

        if type(self.value) == bytes:
            value = base64.standard_b64encode(self.value)
            return "u{:X}:".format(len(value)).encode('ascii') + value

        if type(self.value) == list:
            items = [LiveMessageToken(m).serialize() for m in self.value]
            return b'l' + b''.join(items) + b's'

        if type(self.value) == dict:
            items = []
            for key, value in self.value.items():
                items.append(LiveMessageToken(str(key)).serialize())
                items.append(LiveMessageToken(value).serialize())
            return b'h' + b''.join(items) + b's'

        assert(False)

    @staticmethod
    def deserialize(data):
        def _find(data, needle):
            needle = ord(needle)
            index = 0
            while index < len(data):
                if data[index] == needle:
                    return index
                index += 1
            raise ValueError

        def deserialize_int(data):
            end = _find(data, 's')
            return (LiveMessageToken(int(data[:end], 16)), data[end + 1:])

        def deserialize_list(data):
            result = []
            while data[0] != ord('s'):
                token, data = LiveMessageToken.deserialize(data)
                result.append(token.value)
            return (LiveMessageToken(result), data[1:])

        def deserialize_dict(data):
            result = {}
            while data[0] != ord('s'):
                key, data = LiveMessageToken.deserialize(data)
                value, data = LiveMessageToken.deserialize(data)
                result[str(key.value)] = value.value
            return (LiveMessageToken(result), data[1:])

        def deserialize_string(data, is_base64=False):
            end = _find(data, ':')
            value_end = end + 1 + int(data[:end], 16)
            value = data[end + 1:value_end]
            if is_base64:
                value = base64.standard_b64decode(value)
            else:
                value = value.decode('utf-8')
            return (LiveMessageToken(value), data[value_end:])

        try:
            if data[0] == ord('i'):
                return deserialize_int(data[1:])
            elif data[0] == ord('l'):
                return deserialize_list(data[1:])
            elif data[0] == ord('h'):
                return deserialize_dict(data[1:])
            elif data[0] == ord('u'):
                return deserialize_string(data[1:], is_base64=True)
            else:
                return deserialize_string(data)
        except:
            return (None, None)


class LiveMessage(object):
    def __init__(self, subject=None):
        super(LiveMessage, self).__init__()
        self.tokens = []
        if subject:
            self.append(subject)

    def append(self, value):
        self.tokens.append(value)

    def subject(self):
        return self.tokens[0].lower()

    def parameter(self, index):
        return self.tokens[index + 1]

    def create_signed_message(self, private_key, hash_method):
        data = self.serialize()
        signature = LiveMessage.signature(data, private_key, hash_method)

        envelope = LiveMessage(signature)
        envelope.append(data.decode('utf-8'))
        return envelope

    def verify_signature(self, private_key, hash_method):
        data = self.parameter(0).encode('utf-8')
        signature = LiveMessage.signature(data, private_key, hash_method)
        return self.subject() == signature

    def serialize(self):
        tokens = [LiveMessageToken(t).serialize() for t in self.tokens]
        return b''.join(tokens)

    @staticmethod
    def deserialize(data):
        message = LiveMessage()
        while data:
            token, data = LiveMessageToken.deserialize(data)
            if not token:
                break
            message.append(token.value)
        return message

    @staticmethod
    def signature(data, private_key, hash_method):
        if hash_method == "sha512":
            signature = hashlib.sha512()
        elif hash_method == "sha256":
            signature = hashlib.sha256()
        else:
            signature = hashlib.sha1()

        signature.update(data)
        signature.update(private_key.encode('ascii'))
        return signature.hexdigest().lower()


if __name__ == '__main__':
    msg = LiveMessage('test')
    msg.append([1, 3, 4])
    msg.append({'a': 'A', 'b': 'BB', 'c': 'CCC'})
    msg.append(4)
    msg.append("str")
    msg.append(b'str')
    print(msg.serialize())
    print(LiveMessage.deserialize(msg.serialize()).serialize())
    print(msg.create_signed_message('12345', 'sha1').serialize())
    print(msg.create_signed_message('12345', 'sha1').
          verify_signature('12345', 'sha1'))
